import json

import pytest
import responses

import pjrpc
from pjrpc.client.backend import requests as pjrpc_cli


# TODO refactor tests
@responses.activate
def test_call():
    test_url = 'http://test.com/api'
    responses.add(
        responses.POST, test_url, status=200, json={
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result',
        },
    )

    client = pjrpc_cli.Client(test_url)

    response = client.send(pjrpc.Request('method', (1, 2), id=1))
    assert response.id == 1
    assert response.result == 'result'

    assert responses.calls[0].request.url == test_url
    assert json.loads(responses.calls[0].request.body) == {
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
        'params': [1, 2],
    }

    result = client.call('method', arg1=1, arg2=2)
    assert result == 'result'

    assert responses.calls[1].request.url == test_url
    assert json.loads(responses.calls[1].request.body) == {
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
        'params': {'arg1': 1, 'arg2': 2},
    }

    result = client('method', 1, 2)
    assert result == 'result'

    assert responses.calls[2].request.url == test_url
    assert json.loads(responses.calls[2].request.body) == {
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
        'params': [1, 2],
    }

    result = client.proxy.method(1, 2)
    assert result == 'result'

    assert responses.calls[3].request.url == test_url
    assert json.loads(responses.calls[3].request.body) == {
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
        'params': [1, 2],
    }

    result = client.proxy.method(arg1=1, arg2=2)
    assert result == 'result'

    assert responses.calls[4].request.url == test_url
    assert json.loads(responses.calls[4].request.body) == {
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
        'params': {'arg1': 1, 'arg2': 2},
    }


@responses.activate
def test_notify():
    test_url = 'http://test.com/api'
    responses.add(responses.POST, test_url, status=200, body='')
    client = pjrpc_cli.Client(test_url)

    response = client.send(pjrpc.Request('method', params=[1, 2]))
    assert response is None
    assert responses.calls[0].request.url == test_url
    assert json.loads(responses.calls[0].request.body) == {
        'jsonrpc': '2.0',
        'method': 'method',
        'params': [1, 2],
    }

    response = client.notify('method', 1, 2)
    assert response is None
    assert responses.calls[0].request.url == test_url
    assert json.loads(responses.calls[0].request.body) == {
        'jsonrpc': '2.0',
        'method': 'method',
        'params': [1, 2],
    }


@responses.activate
def test_batch():
    test_url = 'http://test.com/api'
    responses.add(
        responses.POST, test_url, status=200, json=[
            {
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result1',
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'result': 2,
            },
        ],
    )

    client = pjrpc_cli.Client(test_url)

    result = client.batch.send(
        pjrpc.BatchRequest(
            pjrpc.Request('method1', params=[1, 2], id=1),
            pjrpc.Request('method2', params=[2, 3], id=2),
            pjrpc.Request('method3', params=[3, 4]),
        ),
    )
    assert len(result) == 2
    assert result[0].id == 1
    assert result[0].result == 'result1'
    assert result[1].id == 2
    assert result[1].result == 2

    assert responses.calls[0].request.url == test_url
    assert json.loads(responses.calls[0].request.body) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method2',
            'params': [2, 3],
        },
        {
            'jsonrpc': '2.0',
            'method': 'method3',
            'params': [3, 4],
        },
    ]

    result = client.batch[
        ('method1', 1, 2),
        ('method2', 2, 3),
    ]
    assert result == ('result1', 2)

    assert responses.calls[1].request.url == test_url
    assert json.loads(responses.calls[1].request.body) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method2',
            'params': [2, 3],
        },
    ]

    result = client.batch('method1', 1, 2)('method2', 2, 3).call()
    assert result == ('result1', 2)

    assert responses.calls[2].request.url == test_url
    assert json.loads(responses.calls[2].request.body) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method2',
            'params': [2, 3],
        },
    ]

    result = client.batch.proxy.method1(1, 2).method2(2, 3)()
    assert result == ('result1', 2)

    assert responses.calls[3].request.url == test_url
    assert json.loads(responses.calls[3].request.body) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method2',
            'params': [2, 3],
        },
    ]

    result = client.batch.send(
        pjrpc.BatchRequest(
            pjrpc.Request(method='method1', params=[1, 2], id=1),
            pjrpc.Request(method='method2', params=[2, 3], id=2),
        ),
    )
    assert result[0].id == 1
    assert result[0].result == 'result1'
    assert result[1].id == 2
    assert result[1].result == 2

    assert responses.calls[4].request.url == test_url
    assert json.loads(responses.calls[4].request.body) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method2',
            'params': [2, 3],
        },
    ]

    responses.add(responses.POST, test_url, status=200)
    result = client.batch.notify('method1', 1, 2).notify('method2', 2, 3).call()
    assert result is None

    assert responses.calls[5].request.url == test_url
    assert json.loads(responses.calls[5].request.body) == [
        {
            'jsonrpc': '2.0',
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'method': 'method2',
            'params': [2, 3],
        },
    ]


@responses.activate
def test_error():
    test_url = 'http://test.com/api'
    responses.add(
        responses.POST, test_url, status=200, json={
            'jsonrpc': '2.0',
            'id': 1,
            'error': {
                'code': -32601,
                'message': 'Method not found',
            },
        },
    )

    client = pjrpc_cli.Client(test_url)

    with pytest.raises(pjrpc.exceptions.MethodNotFoundError):
        client('method', 1, 2)

    assert responses.calls[0].request.url == test_url
    assert json.loads(responses.calls[0].request.body) == {
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
        'params': [1, 2],
    }

    responses.replace(
        responses.POST, test_url, status=200, json={
            'jsonrpc': '2.0',
            'error': {
                'code': -32600,
                'message': 'Invalid request',
            },
        },
    )

    with pytest.raises(pjrpc.exceptions.InvalidRequestError):
        client.batch[('method', 'param')]

    responses.replace(
        responses.POST, test_url, status=200, json=[
            {
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result1',
            },
            {
                'jsonrpc': '2.0',
                'id': 3,
                'result': 2,
            },
        ],
    )

    client = pjrpc_cli.Client(test_url)

    with pytest.raises(pjrpc.exceptions.IdentityError):
        client.batch[
            ('method1', 'param'),
            ('method2', 'param'),
        ]


@responses.activate
def test_context_manager():
    test_url = 'http://test.com/api'
    responses.add(
        responses.POST, test_url, status=200, json={
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result',
        },
    )

    with pjrpc_cli.Client(test_url) as client:
        response = client.send(pjrpc.Request('method', (1, 2), id=1))
        assert response.id == 1
        assert response.result == 'result'



import pytest
from pjrpc.server import validators


@pytest.mark.parametrize(
    'dyn_method, params', [
        ('param1, param2', [1, 2]),
        ('param1, *args', [1, 2, 3]),
        ('param1, *args', [1]),
        ('param1=1', []),
        ('param1, param2=2', [1]),
        ('param1, param2', {'param1': 1, 'param2': 2}),
        ('param1, **kwargs', {'param1': 1, 'param2': 2, 'param3': 3}),
        ('param1, *, param2, param3', {'param1': 1, 'param2': 2, 'param3': 3}),
        ('param1=1', {}),
        ('param1, param2=2', {'param1': 1}),
        ('param1, *, param2=2', {'param1': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_success(dyn_method, params):
    validator = validators.BaseValidator()
    validator.validate_method(dyn_method, params)


@pytest.mark.parametrize(
    'dyn_method, params', [
        ('param1, param2', [1]),
        ('param1, param2', [1, 2, 3]),
        ('param1, *args', []),
        ('param1, param2', {'param1': 1}),
        ('param1, param2', {'param1': 1, 'param2': 2, 'param3': 3}),
        ('param1, **kwargs', {'param2': 2}),
        ('param1, *, param2, param3', {'param2': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_error(dyn_method, params):
    validator = validators.BaseValidator()

    with pytest.raises(validators.ValidationError):
        validator.validate_method(dyn_method, params)


@pytest.mark.parametrize(
    'dyn_method, exclude, params', [
        ('context, param1', ('context',), [1]),
        ('context, *args', ('context',), []),
        ('context, param1', ('context',), {'param1': 1}),
        ('context, **kwargs', ('context',), {}),
        ('context, *, param1', ('context',), {'param1': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_exclude_success(dyn_method, exclude, params):
    validator = validators.BaseValidator()
    validator.validate_method(dyn_method, params, exclude=exclude)


@pytest.mark.parametrize(
    'dyn_method, exclude, params', [
        ('context, param1', ('context',), [1, 2]),
        ('context, param1', ('context',), {'context': '', 'param1': 1}),
        ('context, *, param1', ('context',), {'context': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_exclude_error(dyn_method, exclude, params):
    validator = validators.BaseValidator()

    with pytest.raises(validators.ValidationError):
        validator.validate_method(dyn_method, params, exclude=exclude)



import pytest

import pjrpc
from pjrpc.common import v20, exceptions


@pytest.mark.parametrize(
    'id, result', [
        (1, None),
        ('1', None),
        (None, None),
        (1, 'result'),
        (1, {'key': 'value'}),
        (1, [1, 2]),
        (1, None),
    ],
)
def test_response_serialization(id, result):
    response = v20.Response(id=id, result=result)

    actual_dict = response.to_json()
    expected_dict = {
        'jsonrpc': '2.0',
        'id': id,
        'result': result,
    }

    assert actual_dict == expected_dict


@pytest.mark.parametrize(
    'id, result', [
        (1, None),
        ('1', None),
        (None, None),
        (1, 'result'),
        (1, {'key': 'value'}),
        (1, [1, 2]),
        (1, None),
    ],
)
def test_response_deserialization(id, result):
    data = {
        'jsonrpc': '2.0',
        'id': id,
        'result': result,
    }

    response = v20.Response.from_json(data)

    assert response.id == id
    assert response.result == result


def test_response_properties():
    response = v20.Response(id=None, result='result')

    assert response.is_success is True
    assert response.is_error is False

    response = v20.Response(id=None, error='error')

    assert response.is_success is False
    assert response.is_error is True


def test_response_repr():
    response = v20.Response(id=1, result='result1')
    assert str(response) == "result1"
    assert repr(response) == "Response(id=1, result='result1', error=UNSET)"


def test_response_deserialization_error():
    with pytest.raises(pjrpc.exc.DeserializationError, match="data must be of type dict"):
        v20.Response.from_json([])

    with pytest.raises(pjrpc.exc.DeserializationError, match="required field 'jsonrpc' not found"):
        v20.Response.from_json({})

    with pytest.raises(pjrpc.exc.DeserializationError, match="jsonrpc version '2.1' is not supported"):
        v20.Response.from_json({'jsonrpc': '2.1'})

    with pytest.raises(pjrpc.exc.DeserializationError, match="field 'id' must be of type integer or string"):
        v20.Response.from_json({'jsonrpc': '2.0', 'id': {}})

    with pytest.raises(pjrpc.exc.DeserializationError, match="'result' or 'error' fields must be provided"):
        v20.Response.from_json({'jsonrpc': '2.0', 'id': 1})

    with pytest.raises(pjrpc.exc.DeserializationError, match="'result' and 'error' fields are mutually exclusive"):
        v20.Response.from_json({'jsonrpc': '2.0', 'id': 1, 'error': {'code': 1, 'message': 'message'}, 'result': 1})


def test_batch_response_serialization():
    response = v20.BatchResponse(
        v20.Response(id=None, result='result0'),
        v20.Response(id=1, result='result1'),
        v20.Response(id=None, result='result2'),
    )

    actual_dict = response.to_json()
    expected_dict = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'result': 'result0',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result1',
        },
        {
            'jsonrpc': '2.0',
            'id': None,
            'result': 'result2',
        },
    ]

    assert actual_dict == expected_dict

    response = v20.BatchResponse(error=pjrpc.exc.MethodNotFoundError())
    actual_dict = response.to_json()
    expected_dict = {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32601,
            'message': 'Method not found',
        },
    }

    assert actual_dict == expected_dict


def test_batch_response_deserialization():
    data = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'result': 'result0',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result1',
        },
        {
            'jsonrpc': '2.0',
            'id': None,
            'result': 'result2',
        },
    ]

    response = v20.BatchResponse.from_json(data)

    assert response[0].id is None
    assert response[0].result == 'result0'

    assert response[1].id == 1
    assert response[1].result == 'result1'

    assert response[2].id is None
    assert response[2].result == 'result2'

    data = {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32601,
            'message': 'Method not found',
        },
    }
    response = v20.BatchResponse.from_json(data)

    assert response.is_error
    assert response.error == pjrpc.exc.MethodNotFoundError()


def test_batch_response_methods():
    response = v20.BatchResponse(
        v20.Response(id=None, result='result0'),
        v20.Response(id=None, result='result0'),
        v20.Response(id=1, result='result1'),
    )
    assert len(response) == 3

    response.append(v20.Response(id=2, result='result2'))
    assert len(response) == 4

    response.extend([
        v20.Response(id=3, result='result3'),
        v20.Response(id=4, result='result4'),
    ])
    assert len(response) == 6

    assert not response.has_error

    response.append(v20.Response(id=5, error=pjrpc.exc.JsonRpcError(code=1, message='msg')))
    assert response.has_error
    with pytest.raises(pjrpc.exc.JsonRpcError) as e:
        response.result

    assert e.value.code == 1
    assert e.value.message == 'msg'


def test_batch_response_errors():
    with pytest.raises(exceptions.IdentityError):
        v20.BatchResponse(
            v20.Response(id=1, result='result1'),
            v20.Response(id=1, result='result1'),
        )


def test_batch_response_repr():
    response = v20.BatchResponse(
        v20.Response(id=None, result='result0'),
        v20.Response(id=1, result='result1'),
    )

    assert str(response) == "[result0, result1]"
    assert repr(response) == "BatchResponse(Response(id=None, result='result0', error=UNSET)," \
                             "Response(id=1, result='result1', error=UNSET), error=UNSET)"


def test_batch_response_deserialization_error():
    with pytest.raises(pjrpc.exc.DeserializationError, match="data must be of type list"):
        v20.BatchResponse.from_json('')



import pytest

import pjrpc
from pjrpc.common import v20, exceptions


@pytest.mark.parametrize(
    'id, params', [
        (1, []),
        ('1', []),
        (None, []),
        (1, None),
        (1, []),
        (1, [1, 2]),
        (1, {'a': 1, 'b': 2}),
    ],
)
def test_request_serialization(id, params):
    request = v20.Request('method', params, id=id)

    actual_dict = request.to_json()
    expected_dict = {
        'jsonrpc': '2.0',
        'method': 'method',
    }
    if id is not None:
        expected_dict.update(id=id)
    if params:
        expected_dict.update(params=params)

    assert actual_dict == expected_dict


@pytest.mark.parametrize(
    'id, params', [
        (1, []),
        ('1', []),
        (None, []),
        (1, None),
        (1, []),
        (1, [1, 2]),
        (1, {'a': 1, 'b': 2}),
    ],
)
def test_request_deserialization(id, params):
    data = {
        'jsonrpc': '2.0',
        'id': id,
        'method': 'method',
    }
    if params:
        data.update(params=params)

    request = v20.Request.from_json(data)

    assert request.id == id
    assert request.method == 'method'
    if params:
        assert request.params == params


def test_request_properties():
    request = v20.Request('method')

    assert request.is_notification is True


def test_request_repr():
    request = v20.Request(method='method', params={'a': 1, 'b': 2})
    assert str(request) == "method(a=1,b=2)"
    assert repr(request) == "Request(method='method', params={'a': 1, 'b': 2}, id=None)"

    request = v20.Request(method='method', params=[1, 2], id=1)
    assert str(request) == "method(1, 2)"
    assert repr(request) == "Request(method='method', params=[1, 2], id=1)"


def test_request_deserialization_error():
    with pytest.raises(pjrpc.exc.DeserializationError, match="data must be of type dict"):
        v20.Request.from_json([])

    with pytest.raises(pjrpc.exc.DeserializationError, match="required field 'jsonrpc' not found"):
        v20.Request.from_json({})

    with pytest.raises(pjrpc.exc.DeserializationError, match="jsonrpc version '2.1' is not supported"):
        v20.Request.from_json({'jsonrpc': '2.1'})

    with pytest.raises(pjrpc.exc.DeserializationError, match="field 'id' must be of type integer or string"):
        v20.Request.from_json({'jsonrpc': '2.0', 'id': {}})

    with pytest.raises(pjrpc.exc.DeserializationError, match="field 'method' must be of type string"):
        v20.Request.from_json({'jsonrpc': '2.0', 'id': 1, 'method': 1})

    with pytest.raises(pjrpc.exc.DeserializationError, match="field 'params' must be of type list or dict"):
        v20.Request.from_json({'jsonrpc': '2.0', 'id': 1, 'method': 'method', 'params': 'params'})


def test_batch_request_serialization():
    request = v20.BatchRequest(
        v20.Request('method0', [], id=None),
        v20.Request('method1', [1, 2], id=1),
        v20.Request('method2', {'a': 1, 'b': 2}, id=None),
    )

    actual_dict = request.to_json()
    expected_dict = [
        {
            'jsonrpc': '2.0',
            'method': 'method0',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'method': 'method2',
            'params': {'a': 1, 'b': 2},
        },
    ]

    assert actual_dict == expected_dict


def test_batch_request_deserialization():
    data = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method0',
            'params': [],
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method2',
            'params': {'a': 1, 'b': 2},
        },
    ]

    request = v20.BatchRequest.from_json(data)

    assert request[0].id is None
    assert request[0].method == 'method0'
    assert request[0].params == []

    assert request[1].id == 1
    assert request[1].method == 'method1'
    assert request[1].params == [1, 2]

    assert request[2].id is None
    assert request[2].method == 'method2'
    assert request[2].params == {'a': 1, 'b': 2}


def test_batch_request_methods():
    request = v20.BatchRequest(
        v20.Request('method1', [1], id=None),
        v20.Request('method2', [1], id=None),
        v20.Request('method3', [1], id=1),
    )
    assert len(request) == 3
    assert not request.is_notification

    request.append(v20.Request('method4', [2], id=2))
    assert len(request) == 4

    request.extend([
        v20.Request('method5', [3], id=3),
        v20.Request('method6', [4], id=4),
    ])
    assert len(request) == 6

    request = v20.BatchRequest(
        v20.Request(id=None, method='method1'),
        v20.Request(id=None, method='method2'),
    )

    assert request.is_notification


def test_batch_request_errors():
    with pytest.raises(exceptions.IdentityError):
        v20.BatchRequest(
            v20.Request(id=1, method='method'),
            v20.Request(id=1, method='method'),
        )


def test_batch_request_repr():
    request = v20.BatchRequest(
        v20.Request('method1', [1, 2]),
        v20.Request('method2', {'a': 1, 'b': 2}, id='2'),
    )

    assert str(request) == "[method1(1, 2), method2(a=1,b=2)]"
    assert repr(request) == "BatchRequest(Request(method='method1', params=[1, 2], id=None)," \
                            "Request(method='method2', params={'a': 1, 'b': 2}, id='2'))"


def test_batch_request_deserialization_error():
    with pytest.raises(pjrpc.exc.DeserializationError, match="data must be of type list"):
        v20.BatchRequest.from_json({})



import json
import yarl
from aioresponses import aioresponses

import pjrpc
from pjrpc.client.backend import aiohttp as pjrpc_cli


# TODO refactor tests
async def test_call():
    test_url = 'http://test.com/api'

    with aioresponses() as m:
        m.post(
            test_url, repeat=True, status=200, payload={
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result',
            },
        )

        client = pjrpc_cli.Client(test_url)

        response = await client.send(pjrpc.Request('method', (1, 2), id=1))

        assert response.id == 1
        assert response.result == 'result'

        assert len(m.requests) == 1
        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[0].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = await client.call('method', arg1=1, arg2=2)
        assert result == 'result'

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[1].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': {'arg1': 1, 'arg2': 2},
        }

        result = await client('method', 1, 2)
        assert result == 'result'

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[2].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = await client.proxy.method(1, 2)
        assert result == 'result'

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[3].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }


async def test_notify():
    test_url = 'http://test.com/api'
    with aioresponses() as m:
        m.post(test_url, repeat=True, status=200, body='')

        client = pjrpc_cli.Client(test_url)

        response = await client.send(pjrpc.Request('method', params=[1, 2]))
        assert response is None
        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[0].kwargs['data']) == {
            'jsonrpc': '2.0',
            'method': 'method',
            'params': [1, 2],
        }

        response = await client.notify('method', a=1, b=2)
        assert response is None
        assert json.loads(requests[1].kwargs['data']) == {
            'jsonrpc': '2.0',
            'method': 'method',
            'params': {'a': 1, 'b': 2},
        }


async def test_batch():
    test_url = 'http://test.com/api'
    with aioresponses() as m:
        m.post(
            test_url, repeat=True, status=200, payload=[
                {
                    'jsonrpc': '2.0',
                    'id': 1,
                    'result': 'result1',
                },
                {
                    'jsonrpc': '2.0',
                    'id': 2,
                    'result': 2,
                },
            ],
        )

        client = pjrpc_cli.Client(test_url)

        result = await client.batch.send(
            pjrpc.BatchRequest(
                pjrpc.Request('method1', params=[1, 2], id=1),
                pjrpc.Request('method2', params=[2, 3], id=2),
                pjrpc.Request('method3', params=[3, 4]),
            ),
        )
        assert len(result) == 2
        assert result[0].id == 1
        assert result[0].result == 'result1'
        assert result[1].id == 2
        assert result[1].result == 2

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[0].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
            {
                'jsonrpc': '2.0',
                'method': 'method3',
                'params': [3, 4],
            },
        ]

        result = await client.batch[
            ('method1', 1, 2),
            ('method2', 2, 3),
        ]
        assert result == ('result1', 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[1].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = await client.batch('method1', 1, 2)('method2', 2, 3).call()
        assert result == ('result1', 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[2].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = await client.batch.proxy.method1(1, 2).method2(2, 3)()
        assert result == ('result1', 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[3].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        m.post(test_url, repeat=True, status=200)
        result = await client.batch.notify('method1', 1, 2).notify('method2', 2, 3).call()
        assert result is None

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[4].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'method': 'method2',
                'params': [2, 3],
            },
        ]


async def test_context_manager():
    test_url = 'http://test.com/api'
    with aioresponses() as m:
        m.post(
            test_url, repeat=True, status=200, payload={
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result',
            },
        )

        async with pjrpc_cli.Client(test_url) as client:
            response = await client.send(pjrpc.Request('method', (1, 2), id=1))

            assert response.id == 1
            assert response.result == 'result'



import pytest
from pjrpc.server import validators
from pjrpc.server.validators import jsonschema


@pytest.mark.parametrize(
    'dyn_method, params, schema', [
        (
            'param1, param2',
            [1, 2],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, *args',
            [1, 2, 3],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'args': {
                        'type': 'array',
                        'items': {
                            'type': 'integer',
                        },
                    },
                },
            },
        ),
        (
            'param1, *args',
            [1],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'args': {
                        'type': 'array',
                        'items': {
                            'type': 'integer',
                        },
                    },
                },
            },
        ),
        (
            'param1=1',
            [],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                },
            },

        ),
        (
            'param1, param2=2',
            [1],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, param2',
            {
                'param1': 1,
                'param2': 2,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, **kwargs',
            {
                'param1': 1,
                'param2': 2,
                'param3': 3,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'kwargs': {'type': 'object'},
                },
            },
        ),
        (
            'param1, *, param2, param3',
            {
                'param1': 1,
                'param2': 2,
                'param3': 3,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                    'param3': {'type': 'integer'},
                },
            },

        ),
        (
            'param1=1',
            {},
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, param2=2',
            {
                'param1': 1,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, *, param2=2',
            {
                'param1': 1,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
    ], indirect=['dyn_method'],
)
def test_validation_success(dyn_method, params, schema):
    validator = jsonschema.JsonSchemaValidator()
    validator.validate_method(dyn_method, params, schema=schema)


@pytest.mark.parametrize(
    'dyn_method, params, schema', [
        (
            'param1, param2',
            [1],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, param2',
            [1, 2, 3],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, *args',
            [],
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'args': {
                        'type': 'array',
                        'items': {
                            'type': 'integer',
                        },
                    },
                },
            },
        ),
        (
            'param1, param2',
            {
                'param1': 1,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, param2',
            {
                'param1': 1,
                'param2': 2,
                'param3': 3,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                },
            },
        ),
        (
            'param1, **kwargs',
            {
                'param2': 2,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'kwargs': {'type': 'object'},
                },
            },
        ),
        (
            'param1, *, param2, param3',
            {
                'param2': 1,
            },
            {
                'type': 'object',
                'properties': {
                    'param1': {'type': 'integer'},
                    'param2': {'type': 'integer'},
                    'param3': {'type': 'integer'},
                },
            },
        ),
    ], indirect=['dyn_method'],
)
def test_validation_error(dyn_method, params, schema):
    validator = jsonschema.JsonSchemaValidator()

    with pytest.raises(validators.ValidationError):
        validator.validate_method(dyn_method, params, schema=schema)



"""
Flask JSON-RPC extension.
"""

from typing import Any

import flask
from flask import current_app
from werkzeug import exceptions

import pjrpc.server


class JsonRPC:
    """
    `Flask <https://flask.palletsprojects.com/en/1.1.x/>`_ framework JSON-RPC extension class.

    :param path: JSON-RPC handler base path
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
    """

    def __init__(self, path: str, **kwargs: Any):
        self._path = path

        kwargs.setdefault('json_loader', flask.json.loads)
        kwargs.setdefault('json_dumper', flask.json.dumps)

        self._dispatcher = pjrpc.server.Dispatcher(**kwargs)

    @property
    def dispatcher(self) -> pjrpc.server.Dispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    def init_app(self, app: flask.Flask) -> None:
        """
        Initializes flask application with JSON-RPC extension.

        :param app: flask application instance
        """

        app.add_url_rule(self._path, methods=['POST'], view_func=self._rpc_handle)

    def _rpc_handle(self) -> flask.Response:
        """
        Handles JSON-RPC request.

        :returns: flask response
        """

        if not flask.request.is_json:
            raise exceptions.UnsupportedMediaType()

        try:
            flask.request.encoding_errors = 'strict'
            request_text = flask.request.get_data(as_text=True)
        except UnicodeDecodeError as e:
            raise exceptions.BadRequest() from e

        response_text = self._dispatcher.dispatch(request_text)
        if response_text is None:
            return current_app.response_class()
        else:
            return current_app.response_class(response_text, mimetype=current_app.config["JSONIFY_MIMETYPE"])



import json

import pjrpc
from pjrpc.server import dispatcher as pjrpc_dispatcher


def test_middleware(mocker):
    test_result = 'the result'
    test_request = pjrpc.common.Request('test_method', params=dict(param='param'), id=1)
    test_response = pjrpc.common.Response(id=1, result=test_result)
    test_context = object()
    middleware_call_order = []

    def test_method(context, param):
        assert context is test_context
        assert param == 'param'
        return test_result

    def test_middleware1(request, context, handler):
        middleware_call_order.append(test_middleware1)
        assert request == test_request
        assert context is test_context

        return handler(request, context)

    def test_middleware2(request, context, handler):
        middleware_call_order.append(test_middleware2)
        assert request == test_request
        assert context is test_context

        return handler(request, context)

    dispatcher = pjrpc_dispatcher.Dispatcher(middlewares=(test_middleware1, test_middleware2))
    dispatcher.add(test_method, 'test_method', 'context')

    request_text = json.dumps(test_request.to_json())
    response_text = dispatcher.dispatch(request_text, test_context)
    actual_response = pjrpc.common.Response.from_json(json.loads(response_text))
    assert actual_response == test_response

    assert middleware_call_order == [test_middleware1, test_middleware2]


async def test_async_middleware(mocker):
    test_result = 'the result'
    test_request = pjrpc.common.Request('test_method', params=dict(param='param'), id=1)
    test_response = pjrpc.common.Response(id=1, result=test_result)
    test_context = object()
    middleware_call_order = []

    async def test_method(context, param):
        assert context is test_context
        assert param == 'param'
        return test_result

    async def test_middleware1(request, context, handler):
        middleware_call_order.append(test_middleware1)
        assert request == test_request
        assert context is test_context

        return await handler(request, context)

    async def test_middleware2(request, context, handler):
        middleware_call_order.append(test_middleware2)
        assert request == test_request
        assert context is test_context

        return await handler(request, context)

    dispatcher = pjrpc_dispatcher.AsyncDispatcher(middlewares=(test_middleware1, test_middleware2))
    dispatcher.add(test_method, 'test_method', 'context')

    request_text = json.dumps(test_request.to_json())
    response_text = await dispatcher.dispatch(request_text, test_context)
    actual_response = pjrpc.common.Response.from_json(json.loads(response_text))
    assert actual_response == test_response

    assert middleware_call_order == [test_middleware1, test_middleware2]



"""
aiohttp JSON-RPC server integration.
"""
from typing import Any, Optional
from aiohttp import web

import pjrpc


class Application:
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param app_args: arguments to be passed to :py:class:`aiohttp.web.Application`
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(self, path: str = '', app: Optional[web.Application] = None, **kwargs: Any):
        self._app = app or web.Application()
        self._dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)

        self._app.router.add_post(path, self.rpc_handle)

    @property
    def app(self) -> web.Application:
        """
        aiohttp application.
        """

        return self._app

    @property
    def dispatcher(self) -> pjrpc.server.Dispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def rpc_handle(self, http_request: web.Request) -> web.Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.content_type != 'application/json':
            raise web.HTTPUnsupportedMediaType()

        try:
            request_text = await http_request.text()
        except UnicodeDecodeError as e:
            raise web.HTTPBadRequest() from e

        response_text = await self._dispatcher.dispatch(request_text, context=http_request)
        if response_text is None:
            return web.Response()
        else:
            return web.json_response(text=response_text)



"""
JSON-RPC client.
"""

from .client import AbstractAsyncClient, AbstractClient
from .tracer import LoggingTracer, Tracer


__all__ = [
    'AbstractClient',
    'AbstractAsyncClient',
    'LoggingTracer',
    'Tracer',
]



import asyncio
import collections
import functools as ft
import json
import logging
from typing import Any, Callable, Dict, Optional, Type, Iterator, Iterable, Union

import pjrpc
from pjrpc.common import v20, BatchRequest, BatchResponse, Request, Response, UNSET, UnsetType
from . import validators

logger = logging.getLogger(__package__)

default_validator = validators.base.BaseValidator()


class Method:
    """
    JSON-RPC method wrapper. Stores method itself and some metainformation.

    :param method: method
    :param name: method name
    :param context: context name
    """

    def __init__(self, method: Callable, name: Optional[str] = None, context: Optional[Any] = None):
        self.method = method
        self.name = name or getattr(method, 'name', method.__name__)
        self.context = context
        self.validator, self.validator_args = getattr(method, '__pjrpc_meta__', (default_validator, {}))

    def bind(self, params: Optional[Union[list, dict]], context: Optional[Any] = None) -> Callable:
        method_params = self.validator.validate_method(
            self.method, params, exclude=(self.context,) if self.context else (), **self.validator_args
        )

        if self.context is not None:
            method_params[self.context] = context

        return ft.partial(self.method, **method_params)


class ViewMethod(Method):
    """
    View method.

    :param view_cls: view class
    :param name: view class method name
    :param context: context name
    """

    def __init__(self, view_cls: Type['ViewMixin'], name: str, context: Optional[Any] = None):
        super().__init__(view_cls, name, context)

        self.view_cls = view_cls

    def bind(self, params: Optional[Union[list, dict]], context: Optional[Any] = None) -> Callable:
        view = self.view_cls(context) if self.context else self.view_cls()
        method = getattr(view, self.name)

        method_params = self.validator.validate_method(method, params, **self.validator_args)

        return ft.partial(method, **method_params)


class ViewMixin:
    """
    Class based method handler mixin.
    """

    @classmethod
    def __methods__(cls):
        for attr_name in filter(lambda name: not name.startswith('_'), dir(cls)):
            attr = getattr(cls, attr_name)
            if callable(attr):
                yield attr


class MethodRegistry:
    """
    Method registry.

    :param prefix: method name prefix to be used for naming containing methods
    """

    def __init__(self, prefix: Optional[str] = None):
        self._prefix = prefix
        self._registry: Dict[str, Method] = {}

    def __iter__(self) -> Iterator[str]:
        """
        Returns registry method iterator.
        """

        return iter(self._registry)

    def __getitem__(self, item: str) -> Method:
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method
        :raises: KeyError
        """

        return self._registry[item]

    def get(self, item: str) -> Optional[Method]:
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method or `None`
        """

        return self._registry.get(item)

    def add(
        self, maybe_method: Optional[Callable] = None, name: Optional[str] = None, context: Optional[Any] = None,
    ) -> Callable:
        """
        Decorator adding decorated method to the registry.

        :param maybe_method: method or `None`
        :param name: method name to be used instead of `__name__` attribute
        :param context: parameter name to be used as an application context
        :returns: decorated method or decorator
        """

        def decorator(method: Callable) -> Callable:
            full_name = '.'.join(filter(None, (self._prefix, name or getattr(method, 'name', method.__name__))))
            self.add_methods(Method(method, full_name, context))

            return method

        if maybe_method is None:
            return decorator
        else:
            return decorator(maybe_method)

    def add_methods(self, *methods: Union[Callable, Method]) -> None:
        """
        Adds methods to the registry.

        :param methods: methods to be added. Each one can be an instance of :py:class:`pjrpc.server.Method`
                        or plain method
        """

        for method in methods:
            if isinstance(method, Method):
                self._add_method(method.name, method)
            else:
                self.add(method)

    def view(
        self, maybe_view: Optional[Type[ViewMixin]] = None, context: Optional[Any] = None, prefix: Optional[str] = None,
    ) -> Union[ViewMixin, Callable]:
        """
        Methods view decorator.

        :param maybe_view: view class instance or `None`
        :param context: application context name
        :param prefix: view methods prefix
        :return: decorator or decorated view
        """

        def decorator(view: Type[ViewMixin]) -> Type[ViewMixin]:
            for method in view.__methods__():
                full_name = '.'.join(filter(None, (self._prefix, prefix, method.__name__)))
                self._add_method(full_name, ViewMethod(view, method.__name__, context))

            return view

        # maybe_view's type depends on the usage of the decorator.  It's a View
        # if it's used as `@view` but ``None`` if used as `@view()`.
        if maybe_view is None:
            return decorator
        else:
            return decorator(maybe_view)

    def merge(self, other: 'MethodRegistry') -> None:
        """
        Merges two registries.

        :param other: registry to be merged in the current one
        """

        for name in other:
            self._add_method(name, other[name])

    def _add_method(self, name: str, method: Method) -> None:
        if name in self._registry:
            logger.warning(f"method '{name}' already registered")

        self._registry[name] = method


class JSONEncoder(pjrpc.JSONEncoder):
    """
    Server JSON encoder. All custom server encoders should be inherited from it.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, validators.base.ValidationError):
            return [err for err in o.args]

        return super().default(o)


class Dispatcher:
    """
    Method dispatcher.

    :param request_class: JSON-RPC request class
    :param response_class: JSON-RPC response class
    :param batch_request: JSON-RPC batch request class
    :param batch_response: JSON-RPC batch response class
    :param json_loader: request json loader
    :param json_dumper: response json dumper
    :param json_encoder: response json encoder
    :param json_decoder: request json decoder
    """

    def __init__(
        self,
        *,
        request_class: Type[Request] = v20.Request,
        response_class: Type[Response] = v20.Response,
        batch_request: Type[BatchRequest] = v20.BatchRequest,
        batch_response: Type[BatchResponse] = v20.BatchResponse,
        json_loader: Callable = json.loads,
        json_dumper: Callable = json.dumps,
        json_encoder: Type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[Type[json.JSONDecoder]] = None,
        middlewares: Iterable[Callable] = (),
    ):
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder
        self._request_class = request_class
        self._response_class = response_class
        self._batch_request = batch_request
        self._batch_response = batch_response
        self._middlewares = list(middlewares)

        self._registry = MethodRegistry()

    def add(self, method: Callable, name: Optional[str] = None, context: Optional[Any] = None) -> None:
        """
        Adds method to the registry.

        :param method: method
        :param name: method name
        :param context: application context name
        """

        self._registry.add(method, name, context)

    def add_methods(self, *methods: Union[MethodRegistry, Method, Callable]) -> None:
        """
        Adds methods to the registry.

        :param methods: method list. Each method may be an instance of :py:class:`pjrpc.server.MethodRegistry`,
                        :py:class:`pjrpc.server.Method` or plain function
        """

        for method in methods:
            if isinstance(method, MethodRegistry):
                self._registry.merge(method)
            elif isinstance(method, Method):
                self._registry.add_methods(method)
            else:
                self._registry.add(method)

    def view(self, view: Type[ViewMixin]) -> None:
        """
        Adds class based view to the registry.

        :param view: view to be added
        """

        self._registry.view(view)

    def dispatch(self, request_text: str, context: Optional[Any] = None) -> Optional[str]:
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation
        """

        logger.debug("request received: %s", request_text)

        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, collections.abc.Iterable):
                response = self._batch_response(
                    *filter(
                        lambda resp: resp is not UNSET, (
                            self._handle_request(request, context) for request in request
                        ),
                    )
                )

            else:
                response = self._handle_request(request, context)

        if response is not UNSET:
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.debug("response sent: %s", response_text)

            return response_text

    def _handle_request(self, request: Request, context: Optional[Any]) -> Union[UnsetType, Response]:
        try:
            handler = self._handle_rpc_request

            for middleware in reversed(self._middlewares):
                handler = ft.partial(middleware, handler=handler)

            return handler(request, context)

        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e

        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, error=error)

    def _handle_rpc_request(self, request: Request, context: Optional[Any]) -> Union[UnsetType, Response]:
        result = self._handle_rpc_method(request.method, request.params, context)
        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, result=result)

    def _handle_rpc_method(self, method_name: str, params: Optional[Union[list, dict]], context: Optional[Any]) -> Any:
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            return method()

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e


class AsyncDispatcher(Dispatcher):
    """
    Asynchronous method dispatcher.
    """

    async def dispatch(self, request_text: str, context: Optional[Any] = None) -> Optional[str]:
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation
        """

        logger.debug("request received: %s", request_text)

        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, collections.Iterable):
                response = self._batch_response(
                    *filter(
                        lambda resp: resp is not UNSET, await asyncio.gather(
                            *(self._handle_request(request, context) for request in request)
                        ),
                    )
                )

            else:
                response = await self._handle_request(request, context)

        if response is not UNSET:
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.debug("response sent: %s", response_text)

            return response_text

    async def _handle_request(self, request: Request, context: Optional[Any]) -> Union[UnsetType, Response]:
        try:
            handler = self._handle_rpc_request

            for middleware in reversed(self._middlewares):
                handler = ft.partial(middleware, handler=handler)

            return await handler(request, context)

        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e

        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, error=error)

    async def _handle_rpc_request(self, request: Request, context: Optional[Any]) -> Union[UnsetType, Response]:
        result = await self._handle_rpc_method(request.method, request.params, context)
        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, result=result)

    async def _handle_rpc_method(
        self, method_name: str, params: Optional[Union[list, dict]], context: Optional[Any],
    ) -> Any:
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            result = method()
            if asyncio.iscoroutine(result):
                result = await result

            return result

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e



import json
from pjrpc.common import BatchRequest, BatchResponse, Request, Response
from pjrpc import client

import pytest


@pytest.mark.parametrize(
    'req, resp, exc', [
        (
            Request(id=1, method='method', params=[1, '2']),
            Response(id=1, result='result'),
            None,
        ),
        (
            BatchRequest(Request(id=1, method='method', params=[1, '2'])),
            BatchResponse(Response(id=1, result='result')),
            None,
        ),
        (
            BatchRequest(Request(id=1, method='method', params=[1, '2'])),
            None,
            BaseException(),
        ),
    ],
)
def test_request_tracing(mocker, req, resp, exc):

    class Client(client.AbstractClient):
        def _request(self, request_text, is_notification=False, **kwargs):
            if exc:
                raise exc
            return json.dumps(resp.to_json())

    class Tracer(client.Tracer):
        on_request_begin = mocker.Mock('on_request_begin')
        on_request_end = mocker.Mock('on_request_end')
        on_error = mocker.Mock('on_error')

    tracer = Tracer()
    cli = Client(tracers=(tracer,))

    trace_ctx = object()

    if exc:
        with pytest.raises(BaseException):
            cli.send(req, _trace_ctx=trace_ctx)
        tracer.on_error.assert_called_once_with(trace_ctx, req, exc)

    else:
        if isinstance(req, BatchRequest):
            cli.batch.send(req, _trace_ctx=trace_ctx)
        else:
            cli.send(req, _trace_ctx=trace_ctx)

        tracer.on_request_begin.assert_called_once_with(trace_ctx, req)
        tracer.on_request_end.assert_called_once_with(trace_ctx, req, resp)


@pytest.mark.parametrize(
    'req, resp, exc', [
        (
            Request(id=1, method='method', params=[1, '2']),
            Response(id=1, result='result'),
            None,
        ),
        (
            BatchRequest(Request(id=1, method='method', params=[1, '2'])),
            BatchResponse(Response(id=1, result='result')),
            None,
        ),
        (
            BatchRequest(Request(id=1, method='method', params=[1, '2'])),
            None,
            BaseException(),
        ),
    ],
)
async def test_async_request_tracing(mocker, req, resp, exc):

    class Client(client.AbstractAsyncClient):
        async def _request(self, request_text, is_notification=False, **kwargs):
            if exc:
                raise exc
            return json.dumps(resp.to_json())

    class Tracer(client.Tracer):
        on_request_begin = mocker.Mock('on_request_begin')
        on_request_end = mocker.Mock('on_request_end')
        on_error = mocker.Mock('on_error')

    tracer = Tracer()
    cli = Client(tracers=(tracer,))

    trace_ctx = object()

    if exc:
        with pytest.raises(BaseException):
            await cli.send(req, _trace_ctx=trace_ctx)
        tracer.on_error.assert_called_once_with(trace_ctx, req, exc)

    else:
        if isinstance(req, BatchRequest):
            await cli.batch.send(req, _trace_ctx=trace_ctx)
        else:
            await cli.send(req, _trace_ctx=trace_ctx)

        tracer.on_request_begin.assert_called_once_with(trace_ctx, req)
        tracer.on_request_end.assert_called_once_with(trace_ctx, req, resp)



import pytest


@pytest.fixture
def dyn_method(request):
    signature = request.param
    context = globals().copy()
    exec(f"def dynamic_method({signature}): pass", context)

    return context['dynamic_method']



"""
Extensible `JSON-RPC <https://www.jsonrpc.org>`_ client/server library.
"""

from pjrpc.common import exceptions
from pjrpc.common import Request, BatchRequest
from pjrpc.common import Response, BatchResponse
from pjrpc.common import JSONEncoder


from pjrpc.__about__ import (
    __title__,
    __description__,
    __url__,
    __version__,
    __author__,
    __email__,
    __license__,
)

# shortcuts
exc = exceptions

__all__ = [
    '__title__',
    '__description__',
    '__url__',
    '__version__',
    '__author__',
    '__email__',
    '__license__',

    'exceptions',
    'exc',
    'Request',
    'Response',
    'BatchRequest',
    'Response',
    'Response',
    'BatchResponse',
    'JSONEncoder',
]
